"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#ProcessorScheduler
	instanceVariableNames:'quiescentProcessLists scheduler zombie activeProcess
		currentPriority readFdArray readSemaphoreArray readCheckArray
		writeFdArray writeSemaphoreArray timeoutArray timeoutActionArray
		timeoutProcessArray timeoutSemaphoreArray idleActions anyTimeouts
		dispatching interruptedProcess useIOInterrupts gotIOInterrupt
		osChildExitActions exitWhenNoMoreUserProcesses'
	classVariableNames:'KnownProcesses KnownProcessIds PureEventDriven
		UserSchedulingPriority UserInterruptPriority TimingPriority
		HighestPriority SchedulingPriority MaxNumberOfProcesses'
	poolDictionaries:''
	category:'Kernel-Processes'
!

!ProcessorScheduler class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    This class has only one instance, which is bound to the global
    'Processor'. It is responsible for scheduling among the smalltalk
    processes (threads; not to confuse with heavy weight unix processes).

    Scheduling is fully done in smalltalk (the always runnable scheduler-
    process, running at highest priority does this).
    The main VM primitive to support this is found in threadSwitch, which passes
    control to another process (usually selected by the scheduler).
    Thus it is possible to modify the schedulers policy and implementation
    at the smalltalk level.
    (To answer a frequently asked question:
     dont add preemptive round-robin here; this can be implemented without
     any need to change the scheduler. See goodies/timeslicing.st for how
     this is done in a very elegant way).

    Notice: Smalltalk/X can (still) be compiled & configured without
    process support. This non-process mode is called 'pureEventDriven' mode
    and is useful to quickly port ST/X to systems, where these facilities
    are either not needed (server applications), or are difficult to
    implement (threads require some assembler support functions). 
    To allow pureEvent mode, kludges are built into some places in the
    system, where either a process is forked, or a timeout is used instead 
    (for examples, see ProcessMonitor or MemoryMonitor).

    This pure-event mode may not be supported in the future 
    (actually, it is no longer maintained in places where was present, so dont
     run the system without Processes).


    class variables:

	KnownProcesses          <Collection>    all known processes
	KnownProcessIds         <Collection>    and their IDs

	PureEventDriven         <Boolean>       true, if no process support
						is available

	UserSchedulingPriority  <Integer>       the priority at which normal
						user interfaces run

	UserInterruptPriority                   the priority at which user-
						interrupts (Cntl-C) processing
						takes place. Processes with
						a greater or equal priority are
						not interruptable.

	TimingPriority                          the priority used for timing.
						Processes with a greater or
						equal priority are not interrupted
						by timers.

	HighestPriority                         The highest allowed prio for processes

	SchedulingPriority                      The priority of the scheduler (must
						me higher than any other).

	MaxNumberOfProcesses                    if non-nil, no more than this
						number of processes are allowed
						(for debugging)

    most interesting methods:

	Processor>>suspend:                  (see also Process>>suspend)
	Processor>>resume:                   (see also Process>>resume)
	Processor>>terminate:                (see also Process>>terminate)
	Processor>>yield 
	Processor>>changePriority:for:       (see also Process>>priority:

	Processor>>signal:afterSeconds:      (see also Delay>>forSeconds:)
	Processor>>signal:afterMilliseconds: (see also Delay>>forMilliseconds:)
	Processor>>signal:onInput:           (see also ExternalStream>>readWait)
	Processor>>signal:onOutput:          (see also ExternalStream>>writeWait)
	Processor>>disableSemaphore:
"
! !

!ProcessorScheduler class methodsFor:'initialization'!

initialize
    "class setup: create the one-and-only instance of myself and
     setup some priority values."

    UserSchedulingPriority := 8.
    UserInterruptPriority := 24.
    TimingPriority := 16.
    SchedulingPriority := 31.
    HighestPriority := 30.

    Processor isNil ifTrue:[
        "create the one and only processor"

        Processor := self basicNew initialize.
    ].

    "
     allow configurations without processes
    "
    PureEventDriven := self threadsAvailable not.
    PureEventDriven ifTrue:[
        'PROCESSOR: no process support - running event driven' errorPrintNL
    ].

    "Modified: 7.3.1996 / 19:22:49 / cg"
! !

!ProcessorScheduler class methodsFor:'instance creation'!

new
    "there is (currently) only one processor ..."

    self error:'only one processor is allowed in the system'
! !

!ProcessorScheduler class methodsFor:'instance release'!

informDispose
    "some Process has been garbage collected 
     - terminate the underlying thread. 
     Usually this does not happen; instead, the process terminates itself 
     by sending #terminate."

    |id sz "{ Class: SmallInteger }"|

    sz := KnownProcessIds size.
    1 to:sz do:[:index |
	(KnownProcesses at:index) isNil ifTrue:[
	    id := KnownProcessIds at:index.
	    id notNil ifTrue:[
		'PROCESSOR: terminating thread ' errorPrint.
		id errorPrint.
		' (no longer refd)' errorPrintNL.

		self threadDestroy:id.
		KnownProcessIds at:index put:nil.
	    ]
	]
    ]
! !

!ProcessorScheduler class methodsFor:'primitive process primitives'!

threadCreate:aProcess withId:id
    "physical creation of a process.
     (warning: low level entry, no administration done).
     This may raise an exception, if a VM process could not be created."

    MaxNumberOfProcesses notNil ifTrue:[
	KnownProcessIds size >= MaxNumberOfProcesses ifTrue:[
	    (KnownProcessIds count:[:el | el notNil]) >= MaxNumberOfProcesses ifTrue:[
		"
		 the number of processes has reached the (soft) limit.
		 This limit prevents runaway programs from creating too many
		 processes. If you continue in the debugger, the process will be
		 created as usual. If you dont want this, abort or terminate.
		"
		self error:'too many processes'.
	    ]
	]
    ].

%{
    int tid;
    extern int __threadCreate();

    tid = __threadCreate(aProcess, 
			 0   /* stackSize: no longer needed */, 
			 __isSmallInteger(id) ? __intVal(id)     /* assign id */
					      : -1              /* let VM assign one */  );
    if (tid) {
	RETURN ( __MKSMALLINT(tid));
    }
%}
.
    "
     arrive here, if creation of process in VM failed.
     This may happen, if the VM does not support more processes,
     or if it ran out of memory, when allocating internal data
     structures.
    "
    ^ ObjectMemory allocationFailureSignal raise.
!

threadDestroy:id
    "physical destroy other process ...
     (warning: low level entry, no administration done)"

%{  /* NOCONTEXT */

    if (__isSmallInteger(id)) {
	__threadDestroy(__intVal(id));
    }
%}
!

threadInterrupt:id
    "make the process evaluate an interrupt. This sets a flag in the VMs
     threadSwitcher, to let the process perform a #interrupt when its set to
     run the next time. The process itself can decide how to react on this 
     interrupt (currently, it looks for interruptBlocks to evaluate)."

%{  /* NOCONTEXT */

    if (__isSmallInteger(id)) {
	__threadInterrupt(__intVal(id));
    }
%}
!

threadsAvailable
    "return true, if the runtime system supports threads (i.e. processes);
     false otherwise."

%{  /* NOCONTEXT */
    RETURN (__threadsAvailable());
%}
! !

!ProcessorScheduler class methodsFor:'queries'!

isPureEventDriven
    "this is temporary - (maybe not :-).
     you can run ST/X either with or without processes.
     Without, there is conceptionally a single process handling all
     outside events and timeouts. This has some negative implications
     (Debugger is ugly), but allows a fully portable ST/X without any
     assembler support - i.e. quick portability.
     The PureEvent flag will automatically be set if the runtime system
     does not support threads - otherwise, it can be set manually
     (from rc-file).
    "

    ^ PureEventDriven
!

knownProcesses
    "return a collection of all (living) processes in the system"

    ^ KnownProcesses select:[:p | p notNil]
!

maxNumberOfProcesses
    "return the limit on the number of processes;
     the default is nil (i.e. unlimited)."

    ^ MaxNumberOfProcesses
!

maxNumberOfProcesses:aNumber
    "set the limit on the number of processes.
     This helps if you have a program which (by error) creates countless
     subprocesses. Without this limit, you may have a hard time to find
     this error (and repairing it). If nil (the default), the number of
     processes is unlimited."

    MaxNumberOfProcesses := aNumber
!

processDriven
    "turn on process driven mode"

    PureEventDriven := false
!

pureEventDriven
    "turn on pure-event driven mode - no processes, single dispatch loop"

    PureEventDriven := true
! !

!ProcessorScheduler methodsFor:'I/O event actions'!

disableFd:aFileDescriptor
    "disable block events on aFileDescriptor.
     This is a leftover support for pure-event systems and may vanish."

    |idx "{Class: SmallInteger }" 
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    idx := readFdArray identityIndexOf:aFileDescriptor startingAt:1.
    idx ~~ 0 ifTrue:[
	readFdArray at:idx put:nil.
	readCheckArray at:idx put:nil.
	readSemaphoreArray at:idx put:nil
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

enableIOAction:aBlock onInput:aFileDescriptor
    "half-obsolete event support: arrange for aBlock to be
     evaluated when input on aFileDescriptor arrives. 
     This is a leftover support for pure-event systems and may vanish."

    |idx "{Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    (readFdArray identityIndexOf:aFileDescriptor startingAt:1) == 0 ifTrue:[
	idx := readFdArray identityIndexOf:nil startingAt:1.
	idx ~~ 0 ifTrue:[
	    readFdArray at:idx put:aFileDescriptor.
	    readCheckArray at:idx put:aBlock.
	    readSemaphoreArray at:idx put:nil
	] ifFalse:[
	    readFdArray := readFdArray copyWith:aFileDescriptor.
	    readCheckArray := readCheckArray copyWith:aBlock.
	    readSemaphoreArray := readSemaphoreArray copyWith:nil.
	]
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
! !

!ProcessorScheduler methodsFor:'accessing'!

activePriority
    "return the priority of the currently running process.
     GNU-ST & ST-80 compatibility; this is the same as currentPriority"

    ^ currentPriority
!

activeProcess
    "return the currently running process"

    ^ activeProcess

    "Processor activeProcess"
!

currentPriority
    "return the priority of the currently running process"

    ^ currentPriority

    "Processor currentPriority"
!

interruptedProcess
    "returns the process which was interrupted by the active one"

    ^ interruptedProcess
! !

!ProcessorScheduler methodsFor:'background processing'!

addIdleBlock:aBlock
    "add the argument, aBlock to the list of idle-actions.
     Idle blocks are evaluated whenever no other process is runnable,
     and no events are pending.
     Use of idle blocks is not recommended, use a low priority processes 
     instead, which has the same effect. Idle blcoks are still included
     to support background actions in pure-event systems, where no processes 
     are available.
     Support for idle-blocks may vanish."

    |wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    idleActions isNil ifTrue:[
	idleActions := OrderedCollection new
    ].
    idleActions add:aBlock.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

removeIdleBlock:aBlock
    "remove the argument, aBlock from the list of idle-blocks.
     Support for idle-blocks may vanish - use low prio processes instead."

    |wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    idleActions notNil ifTrue:[
       idleActions remove:aBlock
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
! !

!ProcessorScheduler methodsFor:'constants'!

highestPriority
    "return the highest priority value (normal) processes can have."

    "must be below schedulingPriority - 
     otherwise scheduler could be blocked ...
    "
    ^ HighestPriority  
!

lowIOPriority
    "not currently used - for ST80 compatibility only"

    ^ 2 "claus: is this ok ?"
!

lowestPriority
    "return the lowest priority value"

    ^ 1   "do not change this - its not variable"
!

schedulingPriority
    "return the priority at which the scheduler runs."

    "must be above highestPriority - 
     otherwise scheduler could be blocked ...
    "
    ^ SchedulingPriority
!

systemBackgroundPriority
    "return the priority, at which background system processing
     should take place.
     Not currently used - for ST80 compatibility only"

    ^ 4
!

timingPriority
    "return the priority, at which all timing takes place (messageTally,
     delay etc.)"

    ^ TimingPriority
!

userBackgroundPriority
    "return the priority, at which background user (non-interactive) processing
     should take place.
     Not currently used - for ST80 compatibility only"

    ^ 6
!

userInterruptPriority
    "return the priority, at which the event scheduler runs - i.e.
     all processes running at a lower priority are interruptable by Cntl-C
     or the timer. Processes running at higher prio will not be interrupted."

    ^ UserInterruptPriority
!

userSchedulingPriority
    "return the priority, at which all normal user (interactive) processing
     takes place"

    ^ UserSchedulingPriority
! !

!ProcessorScheduler methodsFor:'dispatching'!

dispatch
     "It handles timeouts and switches to the highest prio runnable process"

    |any millis pri p nActions "{ Class: SmallInteger }"
     checkBlock sema|

    "
     handle all timeout actions
    "
    anyTimeouts ifTrue:[
	self evaluateTimeouts
    ].

    "first do a quick check for semaphores using checkActions - this is needed for
     devices like the X-connection, where some events might be in the event
     queue. Without these checks, a select might block even though there is work to do
    "
    any := false.
    nActions := readCheckArray size.
    1 to:nActions do:[:index |
	checkBlock := readCheckArray at:index.
	(checkBlock notNil and:[checkBlock value]) ifTrue:[
	    sema := readSemaphoreArray at:index.
	    sema notNil ifTrue:[
		sema signalOnce.
	    ].
	    any := true.
	]
    ].

    "now, someone might be runnable ..."

    p := self highestPriorityRunnableProcess.
    p isNil ifTrue:[
	"/ no one runnable, hard wait for event or timeout

	self waitForEventOrTimeout.
	^ self
    ].

    pri := p priority.

    "
     want to give control to the process p.
     If the switched-to processes priority is lower than the
     userSchedulingPriority, we have to make certain, that the 
     next input or timer will bring us back for a reschedule.
     This is done by enabling ioInterrupts for all file descriptors.
     If ioInterrupts are not available (OS does not support them), 
     we schedule a timer interrupt to interrupt us after 1/20s of a second
     - effectively polling the filedescriptors 20 times a second.
     (which is bad, since low prio processes will be hurt in performance)
     Therefore, dont let benchmarks run with low prio ...

     Higher prio processes must be suspended, 
     same prio ones must yield or suspend to get back control
    "

"
 uncommenting this will make timeouts interrupt the current process
 (i.e. as if the interrupt runs at TimingPrio); 
 if left commented, they are handled at UserSchedulingPrio.
 this will all change, when timeouts are removed and all is process driven
 (a future version will have a process running to handle a timeout queue)
"

"
    pri < TimingPriority ifTrue:[
	anyTimeouts ifTrue:[
	    millis := self timeToNextTimeout.
	    millis == 0 ifTrue:[^ self].
	]
    ].
"

    "
     if the process to run has a lower than UserInterruptPriority,
     arrange for an interrupt to occur on I/O.
     This is done by enabling IO-signals (if the OS supports them)
     or by installing a poll-interrupt after 50ms (if the OS does not).
    "
    pri < UserInterruptPriority ifTrue:[
    
"comment out this if above is uncommented"
	anyTimeouts ifTrue:[
	    millis := self timeToNextTimeout.
	    millis == 0 ifTrue:[^ self].
	].
"---"

	useIOInterrupts ifTrue:[
	    readFdArray do:[:fd |
		fd notNil ifTrue:[
		    OperatingSystem enableIOInterruptsOn:fd
		].
	    ].
	] ifFalse:[
	    millis notNil ifTrue:[
		millis := millis min:50
	    ] ifFalse:[
		millis := 50
	    ]
	]
    ].

    millis notNil ifTrue:[
	"schedule a clock interrupt after millis milliseconds"
	OperatingSystem enableTimer:millis rounded.
    ].

    "
     now let the process run - will come back here by reschedule
     from ioInterrupt or timerInterrupt ... (running at max+1)
    "
    self threadSwitch:p.

    "... when we arrive here, we are back on stage.
	 Either by an ALARM or IO signal, or by a suspend of another process
    "

    millis notNil ifTrue:[
	OperatingSystem disableTimer.
    ].

    "/ check for new input

    (gotIOInterrupt or:[useIOInterrupts not]) ifTrue:[
	gotIOInterrupt := false.
	self checkForInputWithTimeout:0.
    ]

    "Modified: 21.12.1995 / 16:21:54 / stefan"
!

dispatchLoop
    "central dispatch loop; the scheduler process is always staying in
     this method, looping forever."

    |dispatchAction handlerAction ignoredSignals|

    "avoid confusion if entered twice"

    dispatching == true ifTrue:[^ self].
    dispatching := true.

    "/ create the relevant blocks & signalSet outside of the
    "/ while-loop
    "/ (thanks to stefans objectAllocation monitor,
    "/  this safes a bit of memory allocation in the scheduler)

    dispatchAction := [self dispatch].

    handlerAction := [:ex |
			'PROCESSOR: ignored signal' infoPrintNL.
			ex return
		     ].

    ignoredSignals := SignalSet 
			with:(Process terminateSignal)
			with:AbortSignal.

    "/
    "/ I made this an extra call to dispatch; this allows recompilation
    "/  of the dispatch-handling code in the running system.
    "/
    [dispatching] whileTrue:[
	ignoredSignals handle:handlerAction do:dispatchAction
    ].

    "/ we arrive here in standalone Apps,
    "/ when the last process at or above UserSchedulingPriority process died.
    "/ regular ST/X stays in above loop forever

    'PROCESSOR: finish dispatch (no more processes)' infoPrintNL.

    "Modified: 22.12.1995 / 23:12:51 / cg"
!

exitWhenNoMoreUserProcesses:aBoolean
    exitWhenNoMoreUserProcesses := aBoolean
! !

!ProcessorScheduler methodsFor:'os process handling'!

childSignalInterrupt
    "child changed state - execute child termination blocks.
     If child is no longer alive, remove action block."

    |osProcessStatus blocking wasBlocked|

    blocking := OperatingSystem blockingChildProcessWait.

    "/ no interrupt processing, to avoid races with monitorPid
    wasBlocked := OperatingSystem blockInterrupts.
    [
        [
            osProcessStatus := OperatingSystem childProcessWait:blocking.
            osProcessStatus notNil ifTrue:[
                |pid action|

                pid := osProcessStatus pid.
                osProcessStatus stillAlive ifTrue:[
                    action := osChildExitActions at:pid ifAbsent:[].
                ] ifFalse:[
                    action := osChildExitActions removeKey:pid ifAbsent:[].
                ].
                action notNil ifTrue:[
                    action value:osProcessStatus
                ].
            ].

            "/ if pollChildProcesses does block, poll only one status change.
            "/ we will get another SIGCHLD for other status changes.

            osProcessStatus notNil and:[blocking not]
        ] whileTrue.

        "/ if there are no more waiters, disable SIGCHILD handler.
        "/ this helps us with synchronous waiters (e.g. pclose),
        "/ But they should block SIGCHLD anyway.

        osChildExitActions isEmpty ifTrue:[
            OperatingSystem disableChildSignalInterrupts.
        ].
    ] valueNowOrOnUnwindDo:[
        wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
    ]

    "Created: 22.12.1995 / 14:16:26 / stefan"
    "Modified: 5.1.1996 / 16:56:11 / stefan"
    "Modified: 28.2.1996 / 21:36:31 / cg"
!

monitorPid:pid action:aBlock
    "add a 1-arg-block that is called when the operating system child process
     with pid pid changes state.
     The argument for the block is an OSProcessStatus.
    "

    OperatingSystem sigCHLD ~= 0 ifTrue:[
        "/ SIGCHLD is supported,
        "/ aBlock will be evaluated, as soon as a SIGCHLD interrupt for pid has been received.
        OperatingSystem enableChildSignalInterrupts.
        osChildExitActions at:pid put:aBlock
    ] ifFalse:[
        |osProcessStatus|
        "/ SIGCHLD is not supported, wait synchronously for the exit
        "/ of pid.
        [
            osProcessStatus := OperatingSystem childProcessWait:true.
            osProcessStatus notNil ifTrue:[
                (osProcessStatus pid = pid) ifTrue:[
                    aBlock value:osProcessStatus.
                ].
                osProcessStatus stillAlive
            ].
        ] whileTrue.
    ].

    "Created: 28.12.1995 / 14:22:10 / stefan"
    "Modified: 5.1.1996 / 22:01:06 / stefan"
! !

!ProcessorScheduler methodsFor:'primitive process primitives'!

scheduleForInterrupt:aProcess
    "make aProcess evaluate its pushed interrupt block(s)"

    self scheduleInterruptActionsOf:aProcess.
    "
     and, make the process runnable
    "
    aProcess state ~~ #stopped ifTrue:[
	"
	 and, make the process runnable
	"
	aProcess resume
    ]

    "Modified: 5.3.1996 / 17:26:13 / cg"
!

scheduleInterruptActionsOf:aProcess
    "make aProcess evaluate its pushed interrupt block(s)
     when resumed."

    |id|

    aProcess isNil ifTrue:[^ self].
    aProcess == activeProcess ifTrue:[^ self].

    id := aProcess id.
    self class threadInterrupt:id.

    "Created: 5.3.1996 / 17:25:55 / cg"
!

threadSwitch:aProcess
    "continue execution in aProcess.
     (warning: low level entry, no administration is done here)"

    |id pri ok oldProcess oldPri p singleStep wasBlocked|

    (aProcess isNil or:[aProcess == activeProcess]) ifTrue:[^ self].

    wasBlocked := OperatingSystem blockInterrupts.

    oldProcess := activeProcess.
    oldPri := currentPriority.

    id := aProcess id.
    pri := aProcess priority.
    singleStep := aProcess isSingleStepping.
    aProcess state:#active.
    oldProcess setStateTo:#run if:#active.

    "
     no interrupts now - activeProcess has already been changed
     (dont add any message sends here)
    "
    activeProcess := aProcess.
    currentPriority := pri.
%{
    extern OBJ ___threadSwitch();

    if (__isSmallInteger(id)) {
        ok = ___threadSwitch(__context, __intVal(id), (singleStep == true) ? 1 : 0);
    } else {
        ok = false;
    }
%}.
    "time passes spent in some other process ...
     ... here again"

    p := activeProcess.
    activeProcess := oldProcess.
    currentPriority := oldProcess priority.

    ok ifFalse:[
        "
         switch failed for some reason -
         destroy the bad process
        "
        p id ~~ 0 ifTrue:[
            'PROCESSOR: problem with process ' errorPrint. 
                p id errorPrint. 
                p name notNil ifTrue:[
                    ' (' errorPrint. p name errorPrint. ')' errorPrint.
                ].
                '; hard-terminate it.' errorPrintNL.
            p state:#suspended.
            self terminateNoSignal:p.
        ]
    ].
    zombie notNil ifTrue:[
        self class threadDestroy:zombie.
        zombie := nil
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
! !

!ProcessorScheduler methodsFor:'private'!

remember:aProcess
    "remember aProcess for later disposal (where the underlying
     system resources have to be freed)."

    |newShadow oldId wasBlocked
     oldSize "{ Class: SmallInteger }"
     index   "{ Class: SmallInteger }"
     sz      "{ Class: SmallInteger }" |

    wasBlocked := OperatingSystem blockInterrupts.
    index := 1.
    sz := KnownProcessIds size.
    [index <= sz] whileTrue:[
	(KnownProcesses at:index) isNil ifTrue:[
	    oldId := KnownProcessIds at:index.
	    oldId notNil ifTrue:[
		self class threadDestroy:oldId.
	    ].
	    KnownProcesses at:index put:aProcess.
	    KnownProcessIds at:index put:aProcess id.
	    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
	    ^ self
	].
	index := index + 1
    ].

    KnownProcessIds grow:index.
    KnownProcessIds at:index put:aProcess id.

    oldSize := KnownProcesses size.
    (index > oldSize) ifTrue:[
	newShadow := WeakArray new:(oldSize * 2).
	newShadow watcher:self class.
	newShadow replaceFrom:1 with:KnownProcesses.
	KnownProcesses := newShadow
    ].
    KnownProcesses at:index put:aProcess.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

unRemember:aProcess
    "forget aProcess - dispose processing will not consider this one"

    |index wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    index := KnownProcesses identityIndexOf:aProcess.
    index ~~ 0 ifTrue:[
	KnownProcessIds at:index put:nil.
	KnownProcesses at:index put:nil.
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
! !

!ProcessorScheduler methodsFor:'private initializing'!

initialize
    "initialize the one-and-only ProcessorScheduler"

    |nPrios "{ Class: SmallInteger }"
     p|

    KnownProcesses isNil ifTrue:[
	KnownProcesses := WeakArray new:10.
	KnownProcesses watcher:self class.
	KnownProcessIds := OrderedCollection new.
    ].

    "
     create a collection with process lists; accessed using the priority as key
    "
    nPrios := SchedulingPriority.
    quiescentProcessLists := Array new:nPrios.
    1 to:nPrios do:[:pri |
	quiescentProcessLists at:pri put:(LinkedList new)
    ].

    readFdArray := Array with:nil.
    readCheckArray := Array with:nil.
    readSemaphoreArray := Array with:nil.
    writeFdArray := Array with:nil.
    writeSemaphoreArray := Array with:nil.
    timeoutArray := Array with:nil.
    timeoutSemaphoreArray := Array with:nil.
    timeoutActionArray := Array with:nil.
    timeoutProcessArray := Array with:nil.
    anyTimeouts := false.
    dispatching := false.
    exitWhenNoMoreUserProcesses := false. "/ mhmh - how about true ?
    useIOInterrupts := OperatingSystem supportsIOInterrupts.
    gotIOInterrupt := false.
    osChildExitActions := Dictionary new.

    "
     handcraft the first (dispatcher-) process - this one will never
     block, but go into a select if there is nothing to do.
     Also, it has a prio of max+1 - thus, it comes first when looking
     for a runnable process.
    "
    currentPriority := SchedulingPriority.
    p := Process new.
    p setId:0 state:#run.
    p setPriority:currentPriority.
    p name:'scheduler'.

    scheduler := activeProcess := p.

    (quiescentProcessLists at:currentPriority) add:p.

    "
     let me handle IO and timer interrupts
    "
    ObjectMemory ioInterruptHandler:self.
    ObjectMemory timerInterruptHandler:self.
    ObjectMemory childSignalInterruptHandler:self.

    "Modified: 22.12.1995 / 14:15:29 / stefan"
!

reinitialize
    "all previous processes (except those marked as restartable) are made dead 
     - each object should reinstall its process(s) upon restart;
     especially, windowgroups have to.
     In contrast to ST-80, restartable processes are restarted at the beginning
     NOT continued where left. This is a consequence of the portable implementation
     of ST/X, since in order to continue a process, we needed to know the
     internals of the machines (and C-compilers) stack layout.
     This was not done, favouring portability for process continuation.
     In praxis, this is not much of a problem, since in almost every case,
     the computation state can be saved in some object, and processing be 
     restarted from scratch, reinitializing things from this saved state."

    |processesToRestart|

    "
     lay all processes to rest, collect restartable ones
    "
    processesToRestart := OrderedCollection new.
    KnownProcesses do:[:p |
	p notNil ifTrue:[
	    "how, exactly should this be done ?"

	    p isRestartable == true ifTrue:[
		p nextLink:nil.
		processesToRestart add:p
	    ] ifFalse:[
		p setId:nil state:#dead
	    ]
	].
    ].
    scheduler setId:nil state:#dead. 

    "
     now, start from scratch
    "
    KnownProcesses := nil.
    self initialize.

    "
     ... and restart those that can be.
    "
    processesToRestart do:[:p |
"/        'process restart not implemented' errorPrintNL.
	p restart
    ]
! !

!ProcessorScheduler methodsFor:'process creation'!

newProcessFor:aProcess
    "create a physical (VM-) process for aProcess.
     Return true if ok, false if something went wrong.
     The process is not scheduled; to start it running, 
     it needs a Process>>resume. Once resumed, the process will later 
     get control in its #start method."

    |id|

    id := self class threadCreate:aProcess withId:nil.
    id isNil ifTrue:[^ false].

    aProcess setId:id state:#light.   "meaning: has no stack yet"
    self remember:aProcess.
    ^ true
!

newProcessFor:aProcess withId:idWant
    "private entry for Process restart - do not use in your program"

    (self class threadCreate:aProcess withId:idWant) ~~ idWant ifTrue:[
	^ false
    ].

    aProcess state:#light.   "meaning: has no stack yet"
    self remember:aProcess.
    ^ true
! !

!ProcessorScheduler methodsFor:'queries'!

activeProcessIsSystemProcess
    "return true if the active process is a system process,
     which should not be suspended."

    ^ self isSystemProcess:activeProcess

    "
     Processor activeProcessIsSystemProcess
    "
!

anyUserProcessAtAll
    "return true, if there is any process (except myself)
     running at above normal priority.
     This is used to determine if we should stop scheduling
     in standAlone application.
     We ignore background processes to not keep the system from exiting
     solely due to some background GC activity.
     This means:
	You MUST have at least one process running at or above
        UserSchedulingPriority in a standAlone app."

    |listArray l prio "{ Class: SmallInteger }" |

    prio := HighestPriority.
    listArray := quiescentProcessLists.
    [prio >= UserSchedulingPriority] whileTrue:[
	l := listArray at:prio.
	l notEmpty ifTrue:[^ true].
        prio := prio - 1
    ].
    ^ false
!

highestPriorityRunnableProcess
    "return the highest prio runnable process"

    |listArray l p prio "{ Class: SmallInteger }" |

    prio := HighestPriority.
    listArray := quiescentProcessLists.
    [prio >= 1] whileTrue:[
        l := listArray at:prio.
        l notEmpty ifTrue:[
            p := l first.
            "
             if it got corrupted somehow ...
            "
            p id isNil ifTrue:[
                'PROCESSOR: process with nil id removed' errorPrintNL.
                l removeFirst.
                ^ nil.
            ].
            ^ p
        ].
        prio := prio - 1
    ].
    ^ nil

    "Modified: 7.3.1996 / 19:22:05 / cg"
!

isSystemProcess:aProcess
    "return true if aProcess is a system process,
     which should not be suspended/terminated etc.."

    (self class isPureEventDriven 
    or:[aProcess id == 0
    or:[(Display notNil and:[Display dispatchProcess == aProcess])
	" nameOrId endsWith:'dispatcher' "
    ]]) ifTrue:[
	^ true
    ].
    ^ false

    "
     Processor activeProcessIsSystemProcess
    "
! !

!ProcessorScheduler methodsFor:'scheduling'!

changePriority:prio for:aProcess
    "change the priority of aProcess"

    |oldList newList oldPrio newPrio wasBlocked|

    oldPrio := aProcess priority.
    oldPrio == prio ifTrue:[^ self].
    aProcess == scheduler ifTrue:[^ self].

    "
     check for valid argument
    "
    newPrio := prio.
    newPrio < 1 ifTrue:[
        newPrio := 1.
    ] ifFalse:[
        newPrio > HighestPriority ifTrue:[
            newPrio := HighestPriority
        ]
    ].

    [
	wasBlocked := OperatingSystem blockInterrupts.

        aProcess setPriority:newPrio.

        oldList := quiescentProcessLists at:oldPrio.
        (oldList identityIndexOf:aProcess) ~~ 0 ifTrue:[
            oldList remove:aProcess.

            newList := quiescentProcessLists at:newPrio.
            newList addLast:aProcess.

            "if its the current process lowering its prio 
             or another one raising, we have to reschedule"

            aProcess == activeProcess ifTrue:[
                currentPriority := newPrio.
                newPrio < oldPrio ifTrue:[
                    self threadSwitch:scheduler.    
                ]
            ] ifFalse:[
                newPrio > currentPriority ifTrue:[
                    self threadSwitch:aProcess.
                ]
            ].
        ].
    ] valueNowOrOnUnwindDo:[
        wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
    ]

    "Modified: 28.2.1996 / 21:20:47 / cg"
!

interruptActive
    "interrupt the current process"

    activeProcess interrupt
!

processTermination
    "sent by VM if the current process finished its startup block 
     without proper process termination. Lay him to rest now. 
     This can only happen, if something went wrong in Block>>newProcess, 
     since the block defined there always terminates itself."

    self terminateNoSignal:activeProcess.
    self threadSwitch:scheduler
!

reschedule
    "switch to the highest prio runnable process.
     The scheduler itself is always runnable, so we can do an unconditional switch
     to that one. This method is a historical left-over and will vanish."

    ^ self threadSwitch:scheduler
!

resume:aProcess
    "set aProcess runnable - 
     if its prio is higher than the currently running prio, switch to it."

    |l pri wasBlocked|

    (aProcess isNil or:[aProcess == activeProcess]) ifTrue:[^ self].

    "ignore, if process is already dead"
    aProcess id isNil ifTrue:[^ self].

    wasBlocked := OperatingSystem blockInterrupts.

    pri := aProcess priority.

    l := quiescentProcessLists at:pri.
    "if already running, ignore"
    (l identityIndexOf:aProcess) ~~ 0 ifTrue:[
	wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
	^ self
    ].
    l addLast:aProcess.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].

    (pri > currentPriority) ifTrue:[
	"
	 its prio is higher; immediately transfer control to it
	"
	self threadSwitch:aProcess
    ] ifFalse:[
	"
	 its prio is lower; it will have to wait for a while ...
	"
	aProcess state:#run 
    ]
!

resumeForSingleSend:aProcess
    "like resume, but let the process execute a single send only.
     This will be used by the (new, not yet released) debugger 
     for single stepping."

    (aProcess isNil or:[aProcess == activeProcess]) ifTrue:[^ self].
    aProcess singleStep:true.
    self resume:aProcess
!

suspend:aProcess
    "remove the argument, aProcess from the list of runnable processes.
     If the process is the current one, reschedule.
     This method should only be called by Process>>suspend"

    |pri l p wasBlocked|

    "
     some debugging stuff
    "
    aProcess isNil ifTrue:[
	MiniDebugger enterWithMessage:'PROCESSOR: nil suspend'.
	^ self
    ].
    aProcess id isNil ifTrue:[
	MiniDebugger enterWithMessage:'PROCESSOR: bad suspend: already dead'.
	self threadSwitch:scheduler.
	^ self
    ].
    aProcess == scheduler ifTrue:[
	MiniDebugger enterWithMessage:'PROCESSOR: scheduler should never be suspended'.
	^ self
    ].

    wasBlocked := OperatingSystem blockInterrupts.

    pri := aProcess priority.
    l := quiescentProcessLists at:pri.

    "notice: this is slightly faster than putting the if-code into
     the ifAbsent block, because [] is a shared cheap block, created at compile time
    "
    (l remove:aProcess ifAbsent:[]) isNil ifTrue:[
	wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
	'PROCESSOR: bad suspend: not on run list' errorPrintNL.
	"/ MiniDebugger enterWithMessage:'bad suspend: not on run list'.
	self threadSwitch:scheduler.
	^ self
    ].

    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].

    (aProcess == activeProcess) ifTrue:[
	"we can immediately switch sometimes"
	l notEmpty ifTrue:[
	    p := l first
	] ifFalse:[
	    p := scheduler
	].
	self threadSwitch:p 
    ].

    "Modified: 13.12.1995 / 13:32:11 / stefan"
    "Modified: 22.12.1995 / 23:10:12 / cg"
!

terminate:aProcess
    "terminate aProcess. This is donen by sending aProcess the terminateSignal,
     which will evaluate any unwind blocks and finally do a hard terminate."

    aProcess terminate
!

terminateActive
    "terminate the current process (i.e. the running process kills itself).
     The active process is sent the terminateSignal so it will evaluate any
     unwind blocks and finally do a hard terminate.
     This is sent for regular termination and by the VM, if the hard-stack limit
     is reached. (i.e. a process did not repair things in a recursionInterrupt and
     continued to grow its stack)"

    activeProcess terminate
!

terminateActiveNoSignal
    "hard terminate the active process, without sending any
     terminate signal thus no unwind blocks are evaluated."

    self terminateNoSignal:activeProcess
!

terminateNoSignal:aProcess
    "hard terminate aProcess without sending the terminate signal, thus
     no unwind blocks or exitAction are performed in the process.. 
     If its not the current process, it is simply removed from its list 
     and physically destroyed. Otherwise (since we can't take away the chair
     we are sitting on), a switch is forced and the process 
     will be physically destroyed by the next running process. 
     (see zombie handling)"

    |pri id l wasBlocked|

    aProcess isNil ifTrue:[^ self].
    aProcess == scheduler ifTrue:[
	MiniDebugger enterWithMessage:'PROCESSOR: I will not terminate scheduler'.
	^ self
    ].

    id := aProcess id.
    id isNil ifTrue:[^ self].   "already dead"

    aProcess setId:nil state:#dead.

    wasBlocked := OperatingSystem blockInterrupts.

    "remove the process from the runnable list"

    pri := aProcess priority.
    l := quiescentProcessLists at:pri.
    (l identityIndexOf:aProcess) ~~ 0 ifTrue:[
	l remove:aProcess.
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].

    aProcess == activeProcess ifTrue:[
	"
	 hard case - its the currently running process
	 we must have the next active process destroy this one
	 (we cannot destroy the chair we are sitting on ... :-)
	"
	zombie := id.
	self unRemember:aProcess.
	self threadSwitch:scheduler.
	"not reached"
	^ self
    ].
    self class threadDestroy:id.
    self unRemember:aProcess.
    ^ self
!

yield
    "move the currently running process to the end of the currentList
     and reschedule to the first in the list, thus switching to the 
     next same-prio-process."

    |l wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.

    "
     debugging consistency check - will be removed later
    "
    activeProcess priority ~~ currentPriority ifTrue:[
        'PROCESSOR: oops - process changed priority' errorPrintNL.
        currentPriority := activeProcess priority.
    ].

    l := quiescentProcessLists at:currentPriority.

    "
     debugging consistency checks - will be removed later
    "
    l isEmpty ifTrue:[
        wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
        'PROCESSOR: oops - empty runnable list' errorPrintNL.
        ^ self
    ].

    "
     check if the running process is not the only one
    "
    l size ~~ 1 ifTrue:[
        "
         bring running process to the end
        "
        l removeFirst.
        l addLast:activeProcess.

        "
         and switch to first in the list
        "
        self threadSwitch:(l first).
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].

    "Modified: 7.3.1996 / 19:22:43 / cg"
! !

!ProcessorScheduler methodsFor:'semaphore signalling'!

disableSemaphore:aSemaphore
    "disable triggering of a semaphore"

    |idx "{ Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    idx := readSemaphoreArray identityIndexOf:aSemaphore startingAt:1.
    [idx ~~ 0] whileTrue:[
	readFdArray at:idx put:nil.
	readSemaphoreArray at:idx put:nil.
	readCheckArray at:idx put:nil.
	idx := readSemaphoreArray identityIndexOf:aSemaphore startingAt:idx.
    ].
    idx := writeSemaphoreArray identityIndexOf:aSemaphore startingAt:1.
    [idx ~~ 0] whileTrue:[
	writeFdArray at:idx put:nil.
	writeSemaphoreArray at:idx put:nil.
	idx := writeSemaphoreArray identityIndexOf:aSemaphore startingAt:idx.
    ].
    idx := timeoutSemaphoreArray identityIndexOf:aSemaphore startingAt:1.
    [idx ~~ 0] whileTrue:[
	timeoutArray at:idx put:nil.
	timeoutSemaphoreArray at:idx put:nil.
	timeoutActionArray at:idx put:nil.
	timeoutProcessArray at:idx put:nil.
	idx := timeoutSemaphoreArray identityIndexOf:aSemaphore startingAt:idx.
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

signal:aSemaphore afterMilliseconds:millis
    "arrange for a semaphore to be triggered after some milliseconds"

    |now then wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    now := OperatingSystem getMillisecondTime.
    then := OperatingSystem millisecondTimeAdd:now and:millis rounded.
    self signal:aSemaphore atMilliseconds:then.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

signal:aSemaphore afterSeconds:seconds
    "arrange for a semaphore to be triggered after some seconds"

    self signal:aSemaphore afterMilliseconds:(seconds * 1000)
!

signal:aSemaphore atMilliseconds:aMillisecondTime
    "arrange for a semaphore to be triggered at a specific millisecond time.
     If there is already a pending trigger time, the time is changed."

    |index "{ Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    index := timeoutSemaphoreArray identityIndexOf:aSemaphore startingAt:1.
    index ~~ 0 ifTrue:[
	timeoutArray at:index put:aMillisecondTime
    ] ifFalse:[
	index := timeoutArray identityIndexOf:nil startingAt:1.
	index ~~ 0 ifTrue:[
	    timeoutSemaphoreArray at:index put:aSemaphore.
	    timeoutArray at:index put:aMillisecondTime.
	    timeoutActionArray at:index put:nil.
	    timeoutProcessArray at:index put:nil 
	] ifFalse:[
	    timeoutSemaphoreArray := timeoutSemaphoreArray copyWith:aSemaphore.
	    timeoutArray := timeoutArray copyWith:aMillisecondTime.
	    timeoutActionArray := timeoutActionArray copyWith:nil.
	    timeoutProcessArray := timeoutProcessArray copyWith:nil 
	].
    ].

    anyTimeouts := true.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

signal:aSemaphore onInput:aFileDescriptor
    "arrange for a semaphore to be triggered when input on aFileDescriptor
     arrives."

    self signal:aSemaphore onInput:aFileDescriptor orCheck:nil
!

signal:aSemaphore onInput:aFileDescriptor orCheck:aBlock
    "arrange for a semaphore to be triggered when input on aFileDescriptor
     arrives OR checkblock evaluates to true. 
     (checkBlock is used for buffered input, where a select may not detect 
      data already read into a buffer - as in Xlib)"

    |idx "{ Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    (readFdArray identityIndexOf:aFileDescriptor startingAt:1) == 0 ifTrue:[
	idx := readFdArray identityIndexOf:nil startingAt:1.
	idx ~~ 0 ifTrue:[
	    readFdArray at:idx put:aFileDescriptor.
	    readSemaphoreArray at:idx put:aSemaphore.
	    readCheckArray at:idx put:aBlock
	] ifFalse:[
	    readFdArray := readFdArray copyWith:aFileDescriptor.
	    readSemaphoreArray := readSemaphoreArray copyWith:aSemaphore.
	    readCheckArray := readCheckArray copyWith:aBlock.
	]
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

signal:aSemaphore onOutput:aFileDescriptor
    "arrange for a semaphore to be triggered when output on aFileDescriptor
     is possible. (i.e. can be written without blocking)"

    |idx "{ Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    (writeFdArray identityIndexOf:aFileDescriptor startingAt:1) == 0 ifTrue:[
	idx := writeFdArray identityIndexOf:nil startingAt:1.
	idx ~~ 0 ifTrue:[
	    writeFdArray at:idx put:aFileDescriptor.
	    writeSemaphoreArray at:idx put:aSemaphore.
	] ifFalse:[
	    writeFdArray := writeFdArray copyWith:aFileDescriptor.
	    writeSemaphoreArray := writeSemaphoreArray copyWith:aSemaphore.
	]
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
! !

!ProcessorScheduler methodsFor:'timeout handling'!

addTimedBlock:aBlock afterMilliseconds:delta
    "add the argument, aBlock to the list of time-scheduled-blocks; to be
     evaluated after delta milliseconds. The process which installs this timed 
     block will be interrupted for execution of the block.
     (if it is running, the interrupt will occur in whatever method it is
      executing; if it is suspended, it will be resumed).
     The block will be removed from the timed-block list after evaluation 
     (i.e. it will trigger only once)."

    ^ self addTimedBlock:aBlock for:activeProcess afterMilliseconds:delta
!

addTimedBlock:aBlock afterSeconds:delta
    "add the argument, aBlock to the list of time-scheduled-blocks.
     to be evaluated after delta seconds. The process which installs this timed 
     block will be interrupted for execution of the block.
     (if it is running, the interrupt will occur in whatever method it is
      executing; if it is suspended, it will be resumed).
     The block will be removed from the timed-block list after evaluation 
     (i.e. it will trigger only once)."

    self addTimedBlock:aBlock for:activeProcess afterMilliseconds:(delta * 1000) rounded
!

addTimedBlock:aBlock atMilliseconds:aMillisecondTime
    "add the argument, aBlock to the list of time-scheduled-blocks; to be
     evaluated when the millisecondClock value passes aMillisecondTime.
     The process which installs this timed block will be interrupted for 
     execution of the block.
     (if it is running, the interrupt will occur in whatever method it is
      executing; if it is suspended, it will be resumed).
     The block will be removed from the timed-block list after evaluation 
     (i.e. it will trigger only once)."     

    self addTimedBlock:aBlock for:activeProcess atMilliseconds:aMillisecondTime
!

addTimedBlock:aBlock for:aProcess afterMilliseconds:delta
    "add the argument, aBlock to the list of time-scheduled-blocks; to be
     evaluated after delta milliseconds. The process specified by the argument,
     aProcess will be interrupted for execution of the block. 
     (if it is running, the interrupt will occur in whatever method it is
      executing; if it is suspended, it will be resumed).
     If aProcess is nil, the block will be evaluated by the scheduler itself
     (which is dangerous - the block should not raise any error conditions).
     The block will be removed from the timed-block list after evaluation 
     (i.e. it will trigger only once)."

    |now then wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    now := OperatingSystem getMillisecondTime.
    then := OperatingSystem millisecondTimeAdd:now and:delta.
    self addTimedBlock:aBlock for:aProcess atMilliseconds:then.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

addTimedBlock:aBlock for:aProcess afterSeconds:delta
    "add the argument, aBlock to the list of time-scheduled-blocks.
     to be evaluated after delta seconds. aProcess will be interrupted for 
     execution of the block.
     (if it is running, the interrupt will occur in whatever method it is
      executing; if it is suspended, it will be resumed).
     If aProcess is nil, the block will be evaluated by the scheduler itself
     (which is dangerous - the block should not raise any error conditions).
     The block will be removed from the timed-block list after evaluation 
     (i.e. it will trigger only once)."

    self addTimedBlock:aBlock for:aProcess afterMilliseconds:(delta * 1000) rounded
!

addTimedBlock:aBlock for:aProcess atMilliseconds:aMillisecondTime
    "add the argument, aBlock to the list of time-scheduled-blocks; to be
     evaluated by aProcess when the millisecondClock value passes 
     aMillisecondTime. 
     If that block is already in the timeout list, 
     its trigger-time is changed.
     The process specified by the argument, aProcess will be interrupted 
     for execution of the block. 
     If aProcess is nil, the block will be evaluated by the scheduler itself
     (which is dangerous - the block should not raise any error conditions).
     If the process is active at trigger time, the interrupt will occur in 
     whatever method it is executing; if suspended at trigger time, it will be 
     resumed.
     The block will be removed from the timed-block list after evaluation 
     (i.e. it will trigger only once)."     

    |index "{ Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    index := timeoutActionArray identityIndexOf:aBlock startingAt:1.
    index ~~ 0 ifTrue:[
	timeoutArray at:index put:aMillisecondTime
    ] ifFalse:[
	index := timeoutArray indexOf:nil.
	index ~~ 0 ifTrue:[
	    timeoutArray at:index put:aMillisecondTime.
	    timeoutActionArray at:index put:aBlock.
	    timeoutSemaphoreArray at:index put:nil. 
	    timeoutProcessArray at:index put:aProcess 
	] ifFalse:[
	    timeoutArray := timeoutArray copyWith:aMillisecondTime.
	    timeoutActionArray := timeoutActionArray copyWith:aBlock.
	    timeoutSemaphoreArray := timeoutSemaphoreArray copyWith:nil.
	    timeoutProcessArray := timeoutProcessArray copyWith:aProcess.
	].
    ].

    anyTimeouts := true.
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
!

evaluateTimeouts
    "walk through timeouts and evaluate blocks or signal semas that need to be .."

    |sema now aTime block blocksToEvaluate 
     processes n "{ Class: SmallInteger }"|

    anyTimeouts ifFalse:[ ^ self].

    "have to collect the blocks first, then evaluate them. This avoids
     problems due to newly inserted blocks."

    now := OperatingSystem getMillisecondTime.
    blocksToEvaluate := nil.
    n := timeoutArray size.
    anyTimeouts := false.
    1 to:n do:[:index |
	aTime := timeoutArray at:index.
	aTime notNil ifTrue:[
	    (OperatingSystem millisecondTime:aTime isAfter:now) ifFalse:[
		"this one should be triggered"

		sema := timeoutSemaphoreArray at:index.
		sema notNil ifTrue:[
		    sema signalOnce.
		    timeoutSemaphoreArray at:index put:nil
		] ifFalse:[
		    "to support pure-events"
		    block := timeoutActionArray at:index.
		    block notNil ifTrue:[
			blocksToEvaluate isNil ifTrue:[
			    blocksToEvaluate := OrderedCollection new:10.
			    processes := OrderedCollection new:10.
			].
			blocksToEvaluate add:block.
			processes add:(timeoutProcessArray at:index).
			timeoutActionArray at:index put:nil.
			timeoutProcessArray at:index put:nil.
		    ]
		].
		timeoutArray at:index put:nil.
	    ] ifTrue:[
		anyTimeouts := true
	    ]
	]
    ].

    blocksToEvaluate notNil ifTrue:[
	blocksToEvaluate keysAndValuesDo:[:index :block |
	    |p|

	    p := processes at:index.
	    (p isNil or:[p == scheduler or:[PureEventDriven]]) ifTrue:[
		block value
	    ] ifFalse:[
		p interruptWith:block
	    ]
	]
    ]
!

removeTimedBlock:aBlock
    "remove the argument, aBlock from the list of time-sceduled-blocks."

    |index "{ Class: SmallInteger }"
     wasBlocked|

    wasBlocked := OperatingSystem blockInterrupts.
    index := timeoutActionArray identityIndexOf:aBlock startingAt:1.
    (index ~~ 0) ifTrue:[
	timeoutArray at:index put:nil.
	timeoutActionArray at:index put:nil. 
	timeoutSemaphoreArray at:index put:nil.
	timeoutProcessArray at:index put:nil.
    ].
    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
! !

!ProcessorScheduler methodsFor:'waiting'!

checkForInputWithTimeout:millis
    "this is called, when there is absolutely nothing to do;
     hard wait for either input to arrive or a timeout to occur."

    |fd index sema action|

    fd := OperatingSystem 
	      selectOnAnyReadable:readFdArray 
			 writable:writeFdArray
			exception:nil 
		      withTimeOut:millis.

    fd isNil ifTrue:[
	(OperatingSystem lastErrorSymbol == #EBADF) ifTrue:[

	    "/ mhmh - one of the fd's given to me is corrupt.
	    "/ find out which one .... and remove it

	    self removeCorruptedFds
	]
    ] ifFalse:[
	index := readFdArray indexOf:fd.
	index ~~ 0 ifTrue:[
	    sema := readSemaphoreArray at:index.
	    sema notNil ifTrue:[
		sema signalOnce.
		^ true
	    ] ifFalse:[
		action := readCheckArray at:index.
		action notNil ifTrue:[
		    action value.
		     ^ true
		]
	    ]
	]
    ].
    ^ false
!

ioInterrupt
    "data arrived while waiting - switch to scheduler process which will decide 
     what to do now."

    gotIOInterrupt := true.
    interruptedProcess := activeProcess.
    self threadSwitch:scheduler

    "Modified: 21.12.1995 / 16:17:40 / stefan"
!

removeCorruptedFds
    "this is sent when select returns an error due to some invalid 
     fileDescriptor. May happens, if someone does a readWait/writeWait on a 
     connection, which somehow gets corrupted.
     Without special care, all following selects would immediately return with 
     an #EBADF error, leading to high-frequency polling and a locked up system.
     (you could still fix things by interrupting on the console and fixing the
      readFdArray/writeFdArray in the debugger)"

    readFdArray keysAndValuesDo:[:idx :fd |
        |rslt|

        rslt := OperatingSystem
                    selectOnAnyReadable:(Array with:fd)
                               writable:nil
                              exception:nil
                            withTimeOut:0.

        (rslt isNil and:[OperatingSystem lastErrorSymbol == #EBADF]) ifTrue:[
            ('PROCESSOR: remove invalid read fileDescriptor: ' , fd printString) errorPrintNL.
            readFdArray at:idx put:nil
        ]
    ].

    writeFdArray keysAndValuesDo:[:idx :fd |
        |rslt|

        rslt := OperatingSystem
                    selectOnAnyReadable:nil
                               writable:(Array with:fd)
                              exception:nil
                            withTimeOut:0.

        (rslt isNil and:[OperatingSystem lastErrorSymbol == #EBADF]) ifTrue:[
            ('PROCESSOR: removing invalid write fileDescriptor: ' , fd printString) errorPrintNL.
            writeFdArray at:idx put:nil
        ]
    ].
!

schedulerInterrupt
    "forced reschedule - switch to scheduler process which will decide
     what to do now."

    interruptedProcess := activeProcess.
    self threadSwitch:scheduler
!

timeToNextTimeout
    "return the delta-T (in millis) to next timeout, or nil if
     there is none"

    |aTime now delta minDelta n "{ Class: SmallInteger }"|

    "find next timeout. since there are usually not many, just search.
     If there were many, the list should be kept sorted ... keeping deltas
     to next (as in Unix kernel)"

    n := timeoutArray size.
    1 to:n do:[:index |
	aTime := timeoutArray at:index.
	aTime notNil ifTrue:[
	    now isNil ifTrue:[
		now := OperatingSystem getMillisecondTime.
	    ].
	    (OperatingSystem millisecondTime:aTime isAfter:now) ifFalse:[^ 0].
	    delta := OperatingSystem millisecondTimeDeltaBetween:aTime and:now.
	    minDelta isNil ifTrue:[
		minDelta := delta
	    ] ifFalse:[
		minDelta := minDelta min:delta
	    ]
	]
    ].

    ^ minDelta
!

timerInterrupt
    "timer expired while waiting - switch to scheduler process which will decide 
     what to do now."

    interruptedProcess := activeProcess.
    self threadSwitch:scheduler
!

waitForEventOrTimeout
    "entered when no process is runnable - wait for either input on
     any file descriptors to arrive or a timeout to happen.
     If it makes sense, do some background garbage collection.
     The idle actions are a leftover from previous ST/X releases and will
     vanish (installing a low-prio process has the same effect)."

    |millis doingGC|

    doingGC := true.
    [doingGC] whileTrue:[
	anyTimeouts ifTrue:[
	    millis := self timeToNextTimeout.
	    (millis notNil and:[millis <= 0]) ifTrue:[
		^ self    "oops - hurry up checking"
	    ].
	].

	"
	 if its worth doing, collect a bit of garbage;
	 but not, if a backgroundCollector is active
	"
	ObjectMemory backgroundCollectorRunning ifTrue:[
	    doingGC := false
	] ifFalse:[
	    doingGC := ObjectMemory gcStepIfUseful.
	].

	"then do idle actions"
	(idleActions notNil and:[idleActions size ~~ 0]) ifTrue:[
	    idleActions do:[:aBlock |
		aBlock value.
	    ].
	    ^ self   "go back checking"
	].

	doingGC ifTrue:[
	    (self checkForInputWithTimeout:0) ifTrue:[
		^ self  "go back checking"
	    ]
	]
    ].

    "/
    "/ absolutely nothing to do - simply wait
    "/
    exitWhenNoMoreUserProcesses ifTrue:[
	"/ check if there are any processes at all
	"/ stop dispatching if there is none
	"/ (and millis is nil, which means that no timeout blocks are present)
	"/ and no readSemaphores are present (which means that noone is waiting for input)
	"/ and no writeSemaphores are present

	millis isNil ifTrue:[
	    readSemaphoreArray do:[:sema | sema notNil ifTrue:[^ self]].
	    writeSemaphoreArray do:[:sema | sema notNil ifTrue:[^ self]].
	    self anyUserProcessAtAll ifFalse:[
		dispatching := false.
		^ self
	    ]
	]
    ].

    OperatingSystem supportsSelect ifFalse:[
	"SCO instant ShitStation has a bug here,
	 waiting always 1 sec in the select - therefore we delay a bit and
	 return - effectively polling in 50ms cycles
	"
	(self checkForInputWithTimeout:0) ifTrue:[
	    ^ self  "go back checking"
	].
	OperatingSystem millisecondDelay:50.
	^ self
    ].

    millis isNil ifTrue:[
	millis := 9999.
    ] ifFalse:[
	millis := millis rounded
    ].
    self checkForInputWithTimeout:millis

    "Modified: 14.12.1995 / 13:37:46 / stefan"
! !

!ProcessorScheduler class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/ProcessorScheduler.st,v 1.74 1996-04-02 22:04:46 cg Exp $'
! !
ProcessorScheduler initialize!
